@@ -12,38 +12,66 @@
 
 (in-package "SB-VM")
 
+(define-alien-routine ("os_allocate" allocate-system-memory)
+  system-area-pointer
+  (bytes unsigned))
+
+(define-alien-routine ("os_deallocate" deallocate-system-memory)
+  void
+  (addr system-area-pointer)
+  (bytes unsigned))
+
 (!define-load-time-global *allocator-mutex* (sb-thread:make-mutex :name "Allocator"))
 
 (defun allocate-static-vector (widetag length words)
   (declare (type (unsigned-byte #.n-widetag-bits) widetag)
            (type word words)
            (type index length))
-  ;; This does not need WITHOUT-GCING, but it will be implicitly wrapped
-  ;; in WITHOUT-INTERRUPTS due to the default behavior of system mutexes,
-  ;; which is important so that we don't leave an inconsistent state.
-  ;; To think about why it is OK to leave GC enabled, consider that
-  ;; neither GC nor another thread will examine static space above the
-  ;; current value of *STATIC-SPACE-FREE-POINTER*.
-  (or (sb-thread::with-system-mutex (*allocator-mutex*)
-        (let* ((pointer *static-space-free-pointer*)
-               (nbytes (pad-data-block (+ words vector-data-offset)))
-               (new-pointer (sap+ pointer nbytes)))
-          (when (sap<= new-pointer (int-sap static-space-end))
+  ;; Static space starts out zeroed, so it looks a bunch of cons cells
+  ;; containing (0 . 0) above the free pointer. Therefore bumping the pointer
+  ;; merely exposes a new range of cons cells, if GC should happen to run
+  ;; while executing this code. And because we assign the LENGTH of the vector
+  ;; prior to setting the widetag, and LENGTH is a fixnum, then at worst
+  ;; someone can transiently observe a cons of (0 . a-fixnum).
+  (let* ((nbytes (pad-data-block (+ words vector-data-offset)))
+         (pointer (alien-funcall (extern-alien "atomic_bump_static_space_free_ptr"
+                                               (function system-area-pointer int))
+                                 nbytes)))
+    (if (/= (sap-int pointer) 0)
             ;; By storing the length prior to the widetag, the word at the old
             ;; free pointer decodes as a cons instead of a 0-length vector.
             ;; Not that it should matter, but it seems slightly better to change
             ;; the new object atomically to a correctly-sized vector rather than
             ;; a cons changing into the wrong vector into the right vector.
-            (setf (sap-ref-word pointer (ash vector-length-slot word-shift))
-                  (fixnumize length))
+        (let ((v (%make-lisp-obj (sap-int (sap+ pointer other-pointer-lowtag)))))
+          (setf (%array-fill-pointer v) length
             ;; then store the widetag
-            (setf (sap-ref-word pointer 0) widetag)
-            ;; then the new free pointer
-            (setf *static-space-free-pointer* new-pointer)
+                (sap-ref-8 pointer #+big-endian (1- sb-vm:n-word-bytes)
+                                   #+little-endian 0)
+                widetag)
+          v)
+        (error 'simple-storage-condition
+               :format-control
+               "Not enough room left in static space to allocate vector."))))
+
+#+darwin-jit
+(defun allocate-static-code-vector (widetag length words)
+  (declare (type (unsigned-byte #.n-widetag-bits) widetag)
+           (type word words)
+           (type index length))
+  (or (with-system-mutex (*allocator-mutex*)
+        (let* ((pointer *static-code-space-free-pointer*)
+               (nbytes (pad-data-block (+ words vector-data-offset)))
+               (new-pointer (sap+ pointer nbytes)))
+          (when (sap<= new-pointer (int-sap static-code-space-end))
+            (setf (sap-ref-word-jit pointer (ash vector-length-slot word-shift))
+                  (fixnumize length))
+            (setf (sap-ref-word-jit pointer 0) widetag)
+            (setf *static-code-space-free-pointer* new-pointer)
             (%make-lisp-obj (logior (sap-int pointer) other-pointer-lowtag)))))
       (error 'simple-storage-condition
              :format-control
-             "Not enough room left in static space to allocate vector.")))
+             "Not enough room left in static code space to allocate vector.")))
 
 #+immobile-space
 (progn
@@ -196,10 +224,10 @@
     (let ((n-trailing-bytes
            (- (nth-value 1 (ceiling free-ptr immobile-card-bytes)))))
       (setf (sap-ref-word (int-sap free-ptr) 0) simple-array-fixnum-widetag
-            (sap-ref-word (int-sap free-ptr) n-word-bytes)
+            (%array-fill-pointer
+             (%make-lisp-obj (logior free-ptr sb-vm:other-pointer-lowtag)))
             ;; Convert bytes to words, subtract the header and vector length.
-            (ash (- (ash n-trailing-bytes (- word-shift)) 2)
-                 n-fixnum-tag-bits)))))
+            (- (ash n-trailing-bytes (- word-shift)) 2)))))
 
 (defun unallocate (hole)
   #+immobile-space-debug
@@ -273,12 +301,12 @@
       (setf (hole-size hole) (- hole-end hole))))
   (add-to-freelist hole))
 
-(defun allocate-immobile-obj (n-bytes word0 word1 lowtag errorp)
+(defun alloc-immobile-code (n-bytes word0 word1 lowtag errorp)
   (declare (type (and fixnum unsigned-byte) n-bytes))
   (setq n-bytes (align-up n-bytes (* 2 n-word-bytes)))
   ;; Can't allocate fewer than 4 words due to min hole size.
   (aver (>= n-bytes (* 4 n-word-bytes)))
-  (sb-thread::with-system-mutex (*allocator-mutex* :without-gcing t)
+  (with-system-mutex (*allocator-mutex* :without-gcing t)
    (unless (zerop varyobj-holes)
      ;; If deferred sweep needs to happen, do so now.
      ;; Concurrency could potentially be improved here: at most one thread
@@ -319,7 +347,7 @@
                           (sb-debug:print-backtrace)
                           (sb-impl::%halt))
                          (t
-                          (return-from allocate-immobile-obj nil))))
+                          (return-from alloc-immobile-code nil))))
                  (set-varyobj-space-free-pointer free-ptr)
                  addr))))
      (aver (not (logtest addr lowtag-mask))) ; Assert proper alignment
@@ -350,9 +378,8 @@
      (setf (sap-ref-word (int-sap addr) 0) word0
            (sap-ref-word (int-sap addr) n-word-bytes) word1)
      ;; 0-fill the remainder of the object
-     (#+64-bit system-area-ub64-fill
-      #-64-bit system-area-ub32-fill
-      0 (int-sap addr) 2 (- (ash n-bytes (- word-shift)) 2))
+     (alien-funcall (extern-alien "memset" (function void system-area-pointer int unsigned))
+                    (sap+ (int-sap addr) (* 2 n-word-bytes)) 0 (- n-bytes (* 2 n-word-bytes)))
      ;; Only after making the new object can we reduce the size of the hole
      ;; that contained the new allocation (if it entailed chopping a hole
      ;; into parts). In this way, heap scans do not read junk.
@@ -366,34 +393,43 @@
      ;; so that we needn't inhibit GC.
      (%make-lisp-obj (logior addr lowtag)))))
 
-;;; I don't know whether immobile space vectors actually work.
-;;; I think the idea was to use them as fd-stream buffers when applicable
-;;; instead of having a finalizer on OS-allocated memory.
-(defun allocate-immobile-vector (widetag length words)
-  (allocate-immobile-obj (pad-data-block (+ words vector-data-offset))
-                         widetag
-                         (fixnumize length)
-                         other-pointer-lowtag
-                         t))
+;;; Size-class segregation (implying which page we try to allocate to)
+;;; is done from lisp now, not C. There are 4 objects types we'll see,
+;;; each in its own size class (even if some are coincidentally the same size).
+;;;  - Symbols
+;;;  - FDEFNs
+;;;  - Layouts
+;;;  - Funcallable instances
+;;; The first two are truly fixed in size. The latter two could occur in several
+;;; sizes, each occupying a different size-class. This isn't implemented yet.
+(defun alloc-immobile-fixedobj (nwords header)
+  (let* ((widetag (logand (truly-the fixnum header) widetag-mask))
+         (aligned-nwords (truly-the fixnum (align-up nwords 2)))
+         (size-class
+          ;; If you change this, then be sure to update tests/immobile-space.impure
+          ;; which hardcodes a size class to not conflict with anything.
+          (ecase widetag
+            (#.symbol-widetag 1)
+            (#.fdefn-widetag  2)
+            (#.instance-widetag
+             (cond ((<= aligned-nwords  8) (setq aligned-nwords  8) 3)
+                   ((<= aligned-nwords 16) (setq aligned-nwords 16) 4)
+                   ((<= aligned-nwords 24) (setq aligned-nwords 24) 5)
+                   ((<= aligned-nwords 32) (setq aligned-nwords 32) 6)
+                   ((<= aligned-nwords 48) (setq aligned-nwords 48) 7)
+                   (t (error "Oversized layout"))))
+            ;; TODO: allow different sizes of funcallable-instance
+            (#.funcallable-instance-widetag 8))))
+    (values (%primitive alloc-immobile-fixedobj
+                        size-class
+                        aligned-nwords
+                        header))))
 
-(defun allocate-immobile-simple-vector (n-elements)
-  (allocate-immobile-vector simple-vector-widetag n-elements n-elements))
-(defun allocate-immobile-bit-vector (n-elements)
-  (allocate-immobile-vector simple-bit-vector-widetag n-elements
-                            (ceiling n-elements n-word-bits)))
-(defun allocate-immobile-byte-vector (n-elements)
-  (allocate-immobile-vector simple-array-unsigned-byte-8-widetag n-elements
-                            (ceiling n-elements n-word-bytes)))
-(defun allocate-immobile-word-vector (n-elements)
-  (allocate-immobile-vector #+64-bit simple-array-unsigned-byte-64-widetag
-                            #-64-bit simple-array-unsigned-byte-32-widetag
-                            n-elements n-elements))
-
-(defun alloc-immobile-symbol ()
-  (values (%primitive alloc-immobile-fixedobj other-pointer-lowtag symbol-size
-                      (logior (ash (1- symbol-size) n-widetag-bits) symbol-widetag))))
 (defun make-immobile-symbol (name)
-  (let ((symbol (truly-the symbol (alloc-immobile-symbol))))
+  (let ((symbol (truly-the symbol
+                 (alloc-immobile-fixedobj
+                  symbol-size
+                  (logior (ash (1- symbol-size) n-widetag-bits) symbol-widetag)))))
     ;; no pin, it's immobile (and obviously live)
     (setf (sap-ref-lispobj (int-sap (get-lisp-obj-address symbol))
                            (- (ash symbol-name-slot word-shift) other-pointer-lowtag))
@@ -440,10 +476,10 @@
            #+gencgc
            (or #+immobile-code
                (when (member space '(:immobile :auto))
-                 ;; We don't need to inhibit GC here - ALLOCATE-IMMOBILE-OBJ does it.
+                 ;; We don't need to inhibit GC here - ALLOC-IMMOBILE-CODE does it.
                  ;; Indicate that there are initially 2 boxed words, otherwise
                  ;; immobile space GC thinks this object is freeable.
-                 (allocate-immobile-obj (ash total-words word-shift)
+                 (alloc-immobile-code (ash total-words word-shift)
                                         (logior (ash total-words code-header-size-shift)
                                                 code-header-widetag)
                                         (ash 2 n-fixnum-tag-bits)
@@ -477,23 +513,31 @@
         ;; (Could dead immobile objects be converted to use FILLER-WIDETAG instead?)
         (unless (immobile-space-obj-p code)
           ;; Before writing the boxed word count, zeroize up to and including 1 word
-          ;; after the boxed header so that all point words can be safely read
-          ;; by GC and so the jump table count word is 0.
+          ;; after the boxed header so that all pointers can be safely read by GC,
+          ;; and so that the jump table count word is 0.
           (loop for byte-index from (ash boxed word-shift) downto (ash 2 word-shift)
                 by n-word-bytes
-                do (setf (sap-ref-word sap byte-index) 0)))
+                do (setf (sap-ref-word-jit sap byte-index) 0)))
         ;; The 1st slot beyond the header stores the boxed header size in bytes
         ;; as an untagged number, which has the same representation as a tagged
         ;; value denoting a word count if WORD-SHIFT = N-FIXNUM-TAG-BITS.
         ;; This boxed-size MUST be 0 prior to writing any pointers into the object
         ;; because the boxed words will not necessarily have been pre-zeroed;
         ;; scavenging them prior to zeroing them out would see wild pointers.
-        (setf (sap-ref-word sap (ash code-boxed-size-slot word-shift))
+        (setf (sap-ref-word-jit sap (ash code-boxed-size-slot word-shift))
               ;; For 32-bit words, we'll have to add another primitive-object slot.
               ;; But so far nothing makes use of the n-named-calls value.
               (logior #+64-bit (ash n-named-calls 32)
                       (ash boxed word-shift)))))
 
+    ;; FIXME: there may be random values in the unboxed payload and it's not obvious
+    ;; that all callers of ALLOCATE-CODE-OBJECT always write all raw bytes.
+    ;; LOAD-CODE and MAKE-CORE-COMPONENT certainly do because the representation
+    ;; of simple-funs requires that the final 2 bytes be aligned to the physical end
+    ;; of the object so that we can find the function table.
+    ;; But what about other things that create code objects?
+    ;; It could be a subtle source of nondeterministic core images.
+
     ;; FIXME: Sort out 64-bit and cheneygc.
     #+(and 64-bit cheneygc)
     (setf (code-header-ref code 0)
