@@ -13,6 +13,21 @@
 
 (in-package "SB-IMPL")
 
+(macrolet ((def (name &rest args)
+             `(defun ,name ,args
+                (,name ,@args))))
+  (def word-logical-not x)
+  (def word-logical-and x y)
+  (def word-logical-or x y)
+  (def word-logical-xor x y)
+  (def word-logical-nor x y)
+  (def word-logical-eqv x y)
+  (def word-logical-nand x y)
+  (def word-logical-andc1 x y)
+  (def word-logical-andc2 x y)
+  (def word-logical-orc1 x y)
+  (def word-logical-orc2 x y))
+
 (macrolet ((def (name &optional (args '(x)))
              `(defun ,name ,args
                 (,@(if (listp name) `(funcall #',name) `(,name)) ,@args)))
@@ -29,7 +44,8 @@
   (def sap-int)
   (def int-sap)
   (macrolet ((def-accessor (name)
-               `(progn (def ,(symbolicate "%SET-" name) (sap offset value))
+               ;; the low-level %SET functions should not need stubs
+               `(progn (def (setf ,name) (value sap offset))
                        (def ,name (sap offset)))))
     (def-accessor sap-ref-8)
     (def-accessor sap-ref-16)
@@ -46,10 +62,13 @@
     (def-accessor sap-ref-single)
     (def-accessor sap-ref-double))
   (def %byte-blt (src src-start dst dst-start dst-end))
+  (def shift-towards-start (number count))
+  (def shift-towards-end (number count))
   (def get-header-data)
   (def set-header-data (x val))
   (def widetag-of)
   (def %other-pointer-widetag)
+  (def pointer-hash)
   (def vector-sap)
   (def binding-stack-pointer-sap  ())
   ;; x86 uses a plain old inline function for 'dynamic_space_free_pointer'
@@ -57,14 +76,7 @@
   #-(or x86 x86-64) (def dynamic-space-free-pointer ())
   (def control-stack-pointer-sap ())
   (def sb-c:safe-fdefn-fun)
-  (def fun-subtype)
-  (def simple-fun-p)
-  (def %simple-fun-arglist)
-  (def (setf %simple-fun-arglist) (new-value func))
-  (def %simple-fun-name)
-  (def (setf %simple-fun-name) (new-value func))
-  (def %simple-fun-info)
-  (def closurep)
+  (def %fun-pointer-widetag)
   (def %closure-fun)
   (def %closure-index-ref (closure index))
   (def fdefn-name)
@@ -73,8 +85,7 @@
   (def sb-c::vector-length)
   (def make-array-header (type rank))
   (def code-instructions)
-  (def code-header-ref (code-obj index))
-  (def code-header-set (code-obj index new))
+  #-untagged-fdefns (def code-header-ref (code-obj index))
   (def %vector-raw-bits (object offset))
   (def %set-vector-raw-bits (object offset value))
   (def single-float-bits)
@@ -85,26 +96,29 @@
   (def value-cell-ref)
   (def %caller-frame ())
   (def %caller-pc ())
-  (def %code-debug-info)
-  #+(or x86 immobile-space) (def sb-vm::%code-fixups)
+  #+(or x86 x86-64) (def sb-vm::%code-fixups)
+  #+x86-64 (def pointerp)
 
   ;; instances
   (def %make-instance) ; Allocate a new instance with X data slots.
   (def %instance-length) ; Given an instance, return its length.
   (def %instance-layout)
+  (def %instance-wrapper)
   (def %set-instance-layout (instance new-value))
   ; (def %instance-ref (instance index)) ; defined in 'target-defstruct'
   (def %instance-set (instance index new-value))
   ;; funcallable instances
   (def %make-funcallable-instance)
-  (def %funcallable-instance-layout)
-  (def %set-funcallable-instance-layout (fin new-value))
+  (def %fun-layout)
+  (def %fun-wrapper)
+  (def %set-fun-layout (fin new-value))
   (def %funcallable-instance-fun)
   (def (setf %funcallable-instance-fun) (fin new-value))
   (def %funcallable-instance-info (fin i))
   (def %set-funcallable-instance-info (fin i new-value))
-  #+(and compact-instance-header x86-64) (def layout-of)
-  #+64-bit (def layout-depthoid)
+  #+compact-instance-header (progn (def wrapper-of)
+                                   (def %instanceoid-layout))
+
   ;; lists
   (def %rplaca (x val))
   (def %rplacd (x val))
@@ -135,20 +149,53 @@
   (def current-sp ())
   (def current-fp ())
   (def stack-ref (s n))
-  (def %set-stack-ref (s n value))
   (def fun-code-header)
+  (def symbol-hash)
   (def sb-vm::symbol-extra)
-  #+sb-thread (def sb-kernel:symbol-tls-index)
+  #+sb-thread (def symbol-tls-index)
+  #.(if (fboundp 'symbol-info-vector) (values) '(def symbol-info-vector))
   #-(or x86 x86-64) (def lra-code-header)
   (def %make-lisp-obj)
-  (def get-lisp-obj-address))
+  (def get-lisp-obj-address)
+  #+x86-64
+  (def single-float-copysign (float float2))
+  #+x86-64
+  (def single-float-sign))
 
 (defun spin-loop-hint ()
   "Hints the processor that the current thread is spin-looping."
   (spin-loop-hint))
 
+;;; The stub for sb-c::%structure-is-a should really use layout-id in the same way
+;;; that the vop does, however, because the all 64-bit architectures other than
+;;; x86-64 need to use with-pinned-objects to extract a layout-id, it is cheaper not to.
+;;; I should add a vop for uint32 access to raw slots.
+(defun sb-c::%structure-is-a (object-layout test-layout)
+  (or (eq object-layout test-layout)
+      (let ((depthoid (wrapper-depthoid test-layout))
+            (inherits (wrapper-inherits object-layout)))
+        (and (> (length inherits) depthoid)
+             (eq (svref inherits depthoid) test-layout)))))
 
 (defun %other-pointer-subtype-p (x choices)
   (and (%other-pointer-p x)
        (member (%other-pointer-widetag x) choices)
        t))
+
+#+x86-64
+(defun symbol-hash* (x satisfies)
+  (declare (explicit-check)) ; actually, not
+  (declare (ignore satisfies))
+  (symbol-hash* x nil))
+
+;;; TYPECASE could expand to contain a call to this function.
+;;; The interpreter can ignore it, it is just compiler magic.
+(defun sb-c::%type-constraint (var type)
+  (declare (ignore var type))
+  nil)
+(eval-when (:compile-toplevel)
+  ;; Defining %TYPE-CONSTRAINT issues a full warning because TYPE's type
+  ;; is (OR TYPE-SPECIFIER CTYPE), and TYPE-SPECIFIER is
+  ;; (OR LIST SYMBOL CLASSOID CLASS), and CLASS isn't known, and you can't
+  ;; define it because it's a standard symbol.
+  (setq sb-c::*undefined-warnings* nil))
