@@ -148,6 +148,8 @@
   (name (error "missing PACKAGE-DATA-NAME datum"))
   ;; a doc string
   (doc (error "missing PACKAGE-DOC datum"))
+  ;; a list of string designators for shadowing symbols
+  shadow
   ;; a tree containing names for exported symbols which'll be set up at package
   ;; creation time, and NILs, which are ignored. (This is a tree in order to
   ;; allow constructs like '("ENOSPC" #+LINUX ("EDQUOT" "EISNAM" "ENAVAIL"
@@ -190,12 +192,114 @@
     "COERCE" "EXP" "EXPT" "LOG" "SIGNUM" "IMAGPART" "REALPART"
     "ZEROP" "ABS" "SIGNUM" "FLOAT-SIGN"
     "CEILING" "FLOOR" "ROUND" "TRUNCATE" "MOD" "REM"
+    ;; We always want irrational functions to use target floats.
+    "ACOS" "ACOSH" "ASIN" "ASINH" "ATAN" "ATANH"  "CIS" "CONJUGATE"
+    "COS" "COSH"  "FCEILING" "FFLOOR" "FROUND" "FTRUNCATE"
+    "PHASE" "RATIONALIZE" "SIN" "SINH" "SQRT" "TAN" "TANH"
     ;;
     "SXHASH" ; must package-qualify if you mean CL:SXHASH
     ;;
     "BYTE" "BYTE-POSITION" "BYTE-SIZE"
     "DPB" "LDB" "LDB-TEST"
-    "DEPOSIT-FIELD" "MASK-FIELD"))
+    "DEPOSIT-FIELD" "MASK-FIELD"
+    ;;
+    ;; the constants (except for T and NIL which have
+    ;; a specially hacked correspondence between
+    ;; cross-compilation host Lisp and target Lisp)
+
+    ;; We include these here since there is no reason to reference
+    ;; host constants.
+    "ARRAY-DIMENSION-LIMIT"
+    "ARRAY-RANK-LIMIT"
+    "ARRAY-TOTAL-SIZE-LIMIT"
+    "BOOLE-1"
+    "BOOLE-2"
+    "BOOLE-AND"
+    "BOOLE-ANDC1"
+    "BOOLE-ANDC2"
+    "BOOLE-C1"
+    "BOOLE-C2"
+    "BOOLE-CLR"
+    "BOOLE-EQV"
+    "BOOLE-IOR"
+    "BOOLE-NAND"
+    "BOOLE-NOR"
+    "BOOLE-ORC1"
+    "BOOLE-ORC2"
+    "BOOLE-SET"
+    "BOOLE-XOR"
+    "CALL-ARGUMENTS-LIMIT"
+    "CHAR-CODE-LIMIT"
+    "DOUBLE-FLOAT-EPSILON"
+    "DOUBLE-FLOAT-NEGATIVE-EPSILON"
+    "INTERNAL-TIME-UNITS-PER-SECOND"
+    "LAMBDA-LIST-KEYWORDS"
+    "LAMBDA-PARAMETERS-LIMIT"
+    "LEAST-NEGATIVE-DOUBLE-FLOAT"
+    "LEAST-NEGATIVE-LONG-FLOAT"
+    "LEAST-NEGATIVE-NORMALIZED-DOUBLE-FLOAT"
+    "LEAST-NEGATIVE-NORMALIZED-LONG-FLOAT"
+    "LEAST-NEGATIVE-NORMALIZED-SHORT-FLOAT"
+    "LEAST-NEGATIVE-NORMALIZED-SINGLE-FLOAT"
+    "LEAST-NEGATIVE-SHORT-FLOAT"
+    "LEAST-NEGATIVE-SINGLE-FLOAT"
+    "LEAST-POSITIVE-DOUBLE-FLOAT"
+    "LEAST-POSITIVE-LONG-FLOAT"
+    "LEAST-POSITIVE-NORMALIZED-DOUBLE-FLOAT"
+    "LEAST-POSITIVE-NORMALIZED-LONG-FLOAT"
+    "LEAST-POSITIVE-NORMALIZED-SHORT-FLOAT"
+    "LEAST-POSITIVE-NORMALIZED-SINGLE-FLOAT"
+    "LEAST-POSITIVE-SHORT-FLOAT"
+    "LEAST-POSITIVE-SINGLE-FLOAT"
+    "LONG-FLOAT-EPSILON"
+    "LONG-FLOAT-NEGATIVE-EPSILON"
+    "MOST-NEGATIVE-DOUBLE-FLOAT"
+    "MOST-NEGATIVE-FIXNUM"
+    "MOST-NEGATIVE-LONG-FLOAT"
+    "MOST-NEGATIVE-SHORT-FLOAT"
+    "MOST-NEGATIVE-SINGLE-FLOAT"
+    "MOST-POSITIVE-DOUBLE-FLOAT"
+    "MOST-POSITIVE-FIXNUM"
+    "MOST-POSITIVE-LONG-FLOAT"
+    "MOST-POSITIVE-SHORT-FLOAT"
+    "MOST-POSITIVE-SINGLE-FLOAT"
+    "MULTIPLE-VALUES-LIMIT"
+    "PI"
+    "SHORT-FLOAT-EPSILON"
+    "SHORT-FLOAT-NEGATIVE-EPSILON"
+    "SINGLE-FLOAT-EPSILON"
+    "SINGLE-FLOAT-NEGATIVE-EPSILON"
+
+    ;; The cross-compiler itself shouldn't really need to use the host
+    ;; versions of these in target code except in exceptional cases.
+    "CHAR-CODE"
+    "CODE-CHAR"
+    "COMPILE-FILE"
+    "COMPILE-FILE-PATHNAME"
+    "*COMPILE-FILE-PATHNAME*"
+    "*COMPILE-FILE-TRUENAME*"
+    "*COMPILE-PRINT*"
+    "*COMPILE-VERBOSE*"
+    "COMPILER-MACRO-FUNCTION"
+    "CONSTANTP"
+    "GET-SETF-EXPANSION"
+    "*GENSYM-COUNTER*"
+    "LISP-IMPLEMENTATION-TYPE" "LISP-IMPLEMENTATION-VERSION"
+    "MACRO-FUNCTION"
+    "MACROEXPAND" "MACROEXPAND-1" "*MACROEXPAND-HOOK*"
+    "MAKE-LOAD-FORM"
+    "MAKE-LOAD-FORM-SAVING-SLOTS"
+    "PROCLAIM"
+    "SPECIAL-OPERATOR-P"
+    "SUBTYPEP"
+    "UPGRADED-ARRAY-ELEMENT-TYPE"
+    "UPGRADED-COMPLEX-PART-TYPE"
+    "WITH-COMPILATION-UNIT"
+
+    ;; For debugging purposes, we want to be able to intercept inline
+    ;; and block compilation declamations in the host.
+    "DECLAIM"
+    ))
 
 ;;; A symbol in the "dual personality" list refers to the symbol in CL unless
 ;;; package-prefixed with SB-XC:.  The main reason for not putting these
@@ -212,7 +316,14 @@
 ;;;   (* 50 (hash-table-rehash-threshold (make-hash-table)))
 ;;; because we are not intercepting '*.
 (defparameter *dual-personality-math-symbols*
-  '("+" "-" "*" "/" "=" "/=" "<" "<=" ">" ">=" "MIN" "MAX"))
+  '("+" "-" "*" "/" "=" "/=" "<" "<=" ">" ">=" "MIN" "MAX"
+    ;; We've gotten along quite well without an alter-ego for FIXNUM,
+    ;; but now some s-expressions mentioning the type FIXNUM are fed
+    ;; to the host for evaluation, and a type-checking host (such as we are)
+    ;; croaks if an argument exceeds the host's notion of fixnum.
+    ;; Get around it by changing those uses of fixnum to SB-XC:FIXNUM.
+    "FIXNUM"
+    ))
 
 ;;; When playing such tricky package games, it's best to have the symbols that
 ;;; are visible by default, i.e. XC-STRICT-CL:NAME, have no definition,
@@ -224,11 +335,6 @@
 ;;; see by default, so that using them by accident fails.
 (defparameter *undefineds*
   '("SYMBOL-PACKAGE"
-    ;; Irrational functions are never used in the cross-compiler (yet).
-    ;; Prove that by making them undefined.
-    "ACOS" "ACOSH" "ASIN" "ASINH" "ATAN" "ATANH"  "CIS" "CONJUGATE"
-    "COS" "COSH"  "FCEILING" "FFLOOR" "FROUND" "FTRUNCATE"
-    "PHASE" "RATIONALIZE" "SIN" "SINH" "SQRT" "TAN" "TANH"
     ;; Float decoding: don't want to see these used either.
     "DECODE-FLOAT" "INTEGER-DECODE-FLOAT"
     "FLOAT-DIGITS" "FLOAT-PRECISION" "FLOAT-RADIX"
@@ -243,105 +349,17 @@
 ;; compilation of the target.
 ;;
 (let ((package-name "SB-XC"))
-  (dolist (name (append (append *undefineds* *dual-personality-math-symbols*)))
+  (dolist (name (append *undefineds* *dual-personality-math-symbols*))
     (export (intern name package-name) package-name))
-  (dolist (name '(;; the constants (except for T and NIL which have
-                  ;; a specially hacked correspondence between
-                  ;; cross-compilation host Lisp and target Lisp)
-                  "ARRAY-DIMENSION-LIMIT"
-                  "ARRAY-RANK-LIMIT"
-                  "ARRAY-TOTAL-SIZE-LIMIT"
-                  "BOOLE-1"
-                  "BOOLE-2"
-                  "BOOLE-AND"
-                  "BOOLE-ANDC1"
-                  "BOOLE-ANDC2"
-                  "BOOLE-C1"
-                  "BOOLE-C2"
-                  "BOOLE-CLR"
-                  "BOOLE-EQV"
-                  "BOOLE-IOR"
-                  "BOOLE-NAND"
-                  "BOOLE-NOR"
-                  "BOOLE-ORC1"
-                  "BOOLE-ORC2"
-                  "BOOLE-SET"
-                  "BOOLE-XOR"
-                  "CALL-ARGUMENTS-LIMIT"
-                  "CHAR-CODE-LIMIT"
-                  "DOUBLE-FLOAT-EPSILON"
-                  "DOUBLE-FLOAT-NEGATIVE-EPSILON"
-                  "INTERNAL-TIME-UNITS-PER-SECOND"
-                  "LAMBDA-LIST-KEYWORDS"
-                  "LAMBDA-PARAMETERS-LIMIT"
-                  "LEAST-NEGATIVE-DOUBLE-FLOAT"
-                  "LEAST-NEGATIVE-LONG-FLOAT"
-                  "LEAST-NEGATIVE-NORMALIZED-DOUBLE-FLOAT"
-                  "LEAST-NEGATIVE-NORMALIZED-LONG-FLOAT"
-                  "LEAST-NEGATIVE-NORMALIZED-SHORT-FLOAT"
-                  "LEAST-NEGATIVE-NORMALIZED-SINGLE-FLOAT"
-                  "LEAST-NEGATIVE-SHORT-FLOAT"
-                  "LEAST-NEGATIVE-SINGLE-FLOAT"
-                  "LEAST-POSITIVE-DOUBLE-FLOAT"
-                  "LEAST-POSITIVE-LONG-FLOAT"
-                  "LEAST-POSITIVE-NORMALIZED-DOUBLE-FLOAT"
-                  "LEAST-POSITIVE-NORMALIZED-LONG-FLOAT"
-                  "LEAST-POSITIVE-NORMALIZED-SHORT-FLOAT"
-                  "LEAST-POSITIVE-NORMALIZED-SINGLE-FLOAT"
-                  "LEAST-POSITIVE-SHORT-FLOAT"
-                  "LEAST-POSITIVE-SINGLE-FLOAT"
-                  "LONG-FLOAT-EPSILON"
-                  "LONG-FLOAT-NEGATIVE-EPSILON"
-                  "MOST-NEGATIVE-DOUBLE-FLOAT"
-                  "MOST-NEGATIVE-FIXNUM"
-                  "MOST-NEGATIVE-LONG-FLOAT"
-                  "MOST-NEGATIVE-SHORT-FLOAT"
-                  "MOST-NEGATIVE-SINGLE-FLOAT"
-                  "MOST-POSITIVE-DOUBLE-FLOAT"
-                  "MOST-POSITIVE-FIXNUM"
-                  "MOST-POSITIVE-LONG-FLOAT"
-                  "MOST-POSITIVE-SHORT-FLOAT"
-                  "MOST-POSITIVE-SINGLE-FLOAT"
-                  "MULTIPLE-VALUES-LIMIT"
-                  "PI"
-                  "SHORT-FLOAT-EPSILON"
-                  "SHORT-FLOAT-NEGATIVE-EPSILON"
-                  "SINGLE-FLOAT-EPSILON"
-                  "SINGLE-FLOAT-NEGATIVE-EPSILON"
-                  "*READ-DEFAULT-FLOAT-FORMAT*"
-
+  (dolist (name '("*READ-DEFAULT-FLOAT-FORMAT*"
                   "ARRAY-ELEMENT-TYPE"
-                  "CHAR-CODE"
-                  "CODE-CHAR"
-                  "COMPILE-FILE"
-                  "COMPILE-FILE-PATHNAME"
-                  "*COMPILE-FILE-PATHNAME*"
-                  "*COMPILE-FILE-TRUENAME*"
-                  "*COMPILE-PRINT*"
-                  "*COMPILE-VERBOSE*"
-                  "COMPILER-MACRO-FUNCTION"
-                  "CONSTANTP"
-                  "DEFCONSTANT"
-                  "DEFINE-MODIFY-MACRO"
-                  "DEFINE-SETF-EXPANDER"
-                  "DEFMACRO" "DEFSETF" "DEFSTRUCT" "DEFTYPE"
-                  "GENSYM" "*GENSYM-COUNTER*"
-                  "GET-SETF-EXPANSION"
-                  "LISP-IMPLEMENTATION-TYPE" "LISP-IMPLEMENTATION-VERSION"
-                  "MACRO-FUNCTION"
-                  "MACROEXPAND" "MACROEXPAND-1" "*MACROEXPAND-HOOK*"
+                  "DEFMACRO" "DEFSTRUCT" "DEFTYPE"
+                  "GENSYM"
                   "MAKE-ARRAY"
-                  "MAKE-LOAD-FORM"
-                  "MAKE-LOAD-FORM-SAVING-SLOTS"
-                  "PROCLAIM"
                   "SIMPLE-VECTOR"
-                  "SPECIAL-OPERATOR-P"
-                  "SUBTYPEP"
-                  "TYPE-OF" "TYPEP"
-                  "UPGRADED-ARRAY-ELEMENT-TYPE"
-                  "UPGRADED-COMPLEX-PART-TYPE"
-                  "WITH-COMPILATION-UNIT"))
-      (export (intern name package-name) package-name)))
+                  "TYPEP"
+                  ))
+    (export (intern name package-name) package-name)))
 
 (defun count-symbols (pkg)
   (let ((n 0))
@@ -351,7 +369,7 @@
 
 ;;; Build a new package that exports a not-necessarily-strict subset of
 ;;; what the host CL exports. This deals with hosts that have too many
-;;; symbols exported froM CL.
+;;; symbols exported from CL.
 (let ((cl-model-package (make-package "XC-STRICT-CL" :use nil)))
   (flet ((new-external (x package &aux (s (intern x package)))
            (export s package)
@@ -384,8 +402,16 @@
 
 (defun create-target-packages (package-data-list)
   (labels ((flatten (tree)
-             (mapcan (lambda (x) (if (listp x) (flatten x) (list x)))
-                     tree)))
+             (let ((result (mapcan (lambda (x) (if (listp x) (flatten x) (list x)))
+                                   tree)))
+               (when (< (length (remove-duplicates result :test 'equal))
+                        (length result))
+                 (error "Duplicates in package-data-list: ~a~%"
+                        (mapcon (lambda (x)
+                                  (when (member (car x) (cdr x) :test 'equal)
+                                    (list (car x))))
+                                result)))
+               result)))
 
     (hide-host-packages)
 
@@ -394,6 +420,9 @@
     (dolist (package-data package-data-list)
       (let* ((name (package-data-name package-data))
              (package (make-package name :use nil)))
+        ;; Walk the tree of shadowing names
+        (dolist (string (flatten (package-data-shadow package-data)))
+          (shadow string package))
         ;; Walk the tree of exported names, exporting each name.
         (dolist (string (flatten (package-data-export package-data)))
           (export (intern string package) package))))
@@ -450,8 +479,8 @@
           (reexport x))
         (assert (= (length done) (length package-data-list)))))))
 
-(export '*undefined-fun-whitelist*)
-(defvar *undefined-fun-whitelist* (make-hash-table :test 'equal))
+(export '*undefined-fun-allowlist*)
+(defvar *undefined-fun-allowlist* (make-hash-table :test 'equal))
 (let ((list
        (with-open-file (data (prepend-genfile-path "package-data-list.lisp-expr"))
          ;; There's no need to use the precautionary READ-FROM-FILE function
@@ -459,7 +488,7 @@
          (create-target-packages (let ((*readtable* *xc-readtable*)) (read data)))
          (let ((*readtable* *xc-readtable*)) (read data)))))
   (dolist (name (apply #'append list))
-    (setf (gethash name *undefined-fun-whitelist*) t)))
+    (setf (gethash name *undefined-fun-allowlist*) t)))
 
 (defvar *asm-package-use-list*
   '("SB-ASSEM" "SB-DISASSEM"
@@ -491,3 +520,34 @@
             (list (make-package-data :name asm-package
                                      :use (list* "CL" *asm-package-use-list*)
                                      :doc nil)))))
+
+;;; Not all things shown by this are actually unused. Some get removed
+;;; by the tree-shaker as intended.
+#+nil
+(defun show-unused-exports (&aux nonexistent uninteresting)
+  (dolist (entry (with-open-file (f "package-data-list.lisp-expr") (read f)))
+    (let ((pkg (find-package (package-data-name entry))))
+      (dolist (string (mapcan (lambda (x) (if (stringp x) (list x) x))
+                              (package-data-export entry)))
+        (unless (or (string= string "!" :end1 1) (string= string "*!" :end1 2))
+          (let ((s (find-symbol string pkg)))
+            (cond ((not s)
+                   (push (cons pkg string) nonexistent))
+                  ((and (not (boundp s))
+                        (not (sb-kernel:symbol-info s))
+                        (not (gethash s sb-c::*backend-parsed-vops*)))
+                   (push s uninteresting))))))))
+  (format t "~&Nonexistent:~%")
+  (dolist (x nonexistent)
+    (format t "  ~a ~a~%" (package-name (car x)) (cdr x)))
+  (format t "~&Possibly uninteresting:~%")
+  ;; FIXME: prints some things that it shouldn't as "uninteresting"
+  ;; including but not limited to:
+  ;;   - alien struct slot names
+  ;;   - catch tag names (e.g. 'TOPLEVEL-CATCHER)
+  ;;   - declarations
+  ;;   - restart names
+  ;;   - object-not-<type>-error
+  ;;   - markers such as SB-SYS:MACRO (in lexenvs)
+  (dolist (x uninteresting)
+    (format t "  ~s~%" x)))
