@@ -10,46 +10,14 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
+(in-package "SB-VM")
 
 ;;; Make a TN for the argument count passing location for a
 ;;; non-local entry.
 (defun make-nlx-entry-arg-start-location ()
-  (make-wired-tn *fixnum-primitive-type* immediate-arg-scn r8-offset))
+  (make-wired-tn *fixnum-primitive-type* immediate-arg-scn r9-offset))
 
 ;;; Save and restore dynamic environment.
-;;;
-;;; These VOPs are used in the reentered function to restore the appropriate
-;;; dynamic environment.  Currently we only save the Current-Catch and binding
-;;; stack pointer.  We don't need to save/restore the current unwind-protect,
-;;; since unwind-protects are implicitly processed during unwinding.  If there
-;;; were any additional stacks, then this would be the place to restore the top
-;;; pointers.
-
-(define-vop (save-dynamic-state)
-  (:results (catch :scs (descriptor-reg))
-            (nfp :scs (descriptor-reg))
-            (nsp :scs (descriptor-reg)))
-  (:vop-var vop)
-  (:generator 13
-    (load-tl-symbol-value catch *current-catch-block*)
-    (let ((cur-nfp (current-nfp-tn vop)))
-      (when cur-nfp
-        (move nfp cur-nfp)))
-    (inst mov-sp nsp nsp-tn)))
-
-(define-vop (restore-dynamic-state)
-  (:args (catch :scs (descriptor-reg))
-         (nfp :scs (descriptor-reg))
-         (nsp :scs (descriptor-reg)))
-  (:vop-var vop)
-  (:generator 10
-    (store-tl-symbol-value catch *current-catch-block*)
-    (let ((cur-nfp (current-nfp-tn vop)))
-      (when cur-nfp
-        (move cur-nfp nfp)))
-    (inst mov-sp nsp-tn nsp)))
-
 (define-vop (current-stack-pointer)
   (:results (res :scs (any-reg descriptor-reg)))
   (:generator 1
@@ -80,17 +48,30 @@
   (:info entry-label)
   (:results (block :scs (any-reg)))
   (:temporary (:scs (descriptor-reg)) temp)
-  (:temporary (:scs (interior-reg)) lip)
+  (:vop-var vop)
   (:generator 22
-    (inst add block cfp-tn (add-sub-immediate (* (tn-offset tn) n-word-bytes)))
+    (inst add block cfp-tn (add-sub-immediate (tn-byte-offset tn)))
     (load-tl-symbol-value temp *current-unwind-protect-block*)
-    #.(assert (and (= unwind-block-uwp-slot 0)
-                   (= unwind-block-cfp-slot 1)))
-    (inst stp temp cfp-tn (@ block))
-    (inst compute-lra temp lip entry-label)
-    #.(assert (and (= unwind-block-code-slot 2)
-                   (= unwind-block-entry-pc-slot 3)))
-    (inst stp code-tn temp (@ block (* n-word-bytes 2)))))
+    (storew-pair temp unwind-block-uwp-slot cfp-tn unwind-block-cfp-slot block)
+    (inst adr temp entry-label)
+    (storew temp block unwind-block-entry-pc-slot)
+    #+sb-thread
+    (loadw-pair
+     temp (/ (info :variable :wired-tls '*binding-stack-pointer*) n-word-bytes)
+     tmp-tn (/ (info :variable :wired-tls '*current-catch-block*) n-word-bytes)
+     thread-tn)
+    #-sb-thread
+    (progn
+      (load-binding-stack-pointer temp)
+      (load-tl-symbol-value tmp-tn *current-catch-block*))
+    (storew-pair temp unwind-block-bsp-slot tmp-tn unwind-block-current-catch-slot block)
+    (inst mov-sp temp nsp-tn)
+    (let ((nfp (current-nfp-tn vop)))
+      (if nfp
+          (storew-pair nfp unwind-block-nfp-slot
+                       temp unwind-block-nsp-slot
+                       block)
+          (storew temp block unwind-block-nsp-slot)))))
 
 ;;; Like Make-Unwind-Block, except that we also store in the specified tag, and
 ;;; link the block into the Current-Catch list.
@@ -100,25 +81,33 @@
   (:info entry-label)
   (:results (block :scs (any-reg)))
   (:temporary (:scs (descriptor-reg)) temp)
-  (:temporary (:scs (descriptor-reg) :target block :to (:result 0)) result)
-  (:temporary (:scs (interior-reg)) lip)
+  (:vop-var vop)
   (:generator 44
-    (inst add result cfp-tn (add-sub-immediate (* (tn-offset tn) n-word-bytes)))
+    (inst add block cfp-tn (add-sub-immediate (tn-byte-offset tn)))
     (load-tl-symbol-value temp *current-unwind-protect-block*)
-    #.(assert (and (= catch-block-uwp-slot 0)
-                   (= catch-block-cfp-slot 1)))
-    (inst stp temp cfp-tn (@ result))
-    #.(assert (and (= catch-block-code-slot 2)
-                   (= catch-block-entry-pc-slot 3)))
-    (inst compute-lra temp lip entry-label)
-    (inst stp code-tn temp (@ result (* n-word-bytes 2)))
-    #.(assert (and (= catch-block-previous-catch-slot 4)
-                   (= catch-block-tag-slot 5)))
-    (load-tl-symbol-value temp *current-catch-block*)
-    (inst stp temp tag (@ result (* n-word-bytes 4)))
-    (store-tl-symbol-value result *current-catch-block*)
+    (storew-pair temp catch-block-uwp-slot cfp-tn catch-block-cfp-slot block)
+    (inst adr temp entry-label)
+    (storew temp block catch-block-entry-pc-slot)
 
-    (move block result)))
+    #+sb-thread
+    (loadw-pair
+     temp (/ (info :variable :wired-tls '*binding-stack-pointer*) n-word-bytes)
+     tmp-tn (/ (info :variable :wired-tls '*current-catch-block*) n-word-bytes)
+     thread-tn)
+    #-sb-thread
+    (progn
+      (load-binding-stack-pointer temp)
+      (load-tl-symbol-value tmp-tn *current-catch-block*))
+    (storew-pair tmp-tn catch-block-previous-catch-slot tag catch-block-tag-slot block)
+    (storew temp block catch-block-bsp-slot)
+    (inst mov-sp temp nsp-tn)
+    (let ((nfp (current-nfp-tn vop)))
+      (if nfp
+          (storew-pair nfp unwind-block-nfp-slot
+                       temp unwind-block-nsp-slot
+                       block)
+          (storew temp block unwind-block-nsp-slot)))
+    (store-tl-symbol-value block *current-catch-block*)))
 
 ;;; Just set the current unwind-protect to UWP.  This
 ;;; instantiates an unwind block as an unwind-protect.
@@ -127,19 +116,21 @@
   (:generator 7
     (store-tl-symbol-value uwp *current-unwind-protect-block*)))
 
-(define-vop (unlink-catch-block)
+(define-vop (%catch-breakup)
+  (:args (current-block))
+  (:ignore current-block)
   (:temporary (:scs (any-reg)) block)
   (:policy :fast-safe)
-  (:translate %catch-breakup)
   (:generator 17
     (load-tl-symbol-value block *current-catch-block*)
     (loadw block block catch-block-previous-catch-slot)
     (store-tl-symbol-value block *current-catch-block*)))
 
-(define-vop (unlink-unwind-protect)
+(define-vop (%unwind-protect-breakup)
+  (:args (current-block))
+  (:ignore current-block)
   (:temporary (:scs (any-reg)) block)
   (:policy :fast-safe)
-  (:translate %unwind-protect-breakup)
   (:generator 17
     (load-tl-symbol-value block *current-unwind-protect-block*)
     (loadw block block unwind-block-uwp-slot)
@@ -158,7 +149,7 @@
   (:save-p :force-to-stack)
   (:vop-var vop)
   (:generator 30
-    (emit-return-pc label)
+    (emit-label label)
     (note-this-location vop :non-local-entry)
     (cond ((zerop nvals))
           ((= nvals 1)
@@ -204,7 +195,7 @@
   (:save-p :force-to-stack)
   (:vop-var vop)
   (:generator 30
-    (emit-return-pc label)
+    (emit-label label)
     (note-this-location vop :non-local-entry)
 
     ;; Setup results, and test for the zero value case.
@@ -236,20 +227,33 @@
 (define-vop (uwp-entry)
   (:info label)
   (:save-p :force-to-stack)
-  (:results (block) (start) (count))
-  (:ignore block start count)
   (:vop-var vop)
   (:generator 0
-    (emit-return-pc label)
+    (emit-label label)
     (note-this-location vop :non-local-entry)))
 
+(define-vop (uwp-entry-block)
+  (:info label)
+  (:save-p :force-to-stack)
+  (:results (block))
+  (:vop-var vop)
+  (:generator 0
+    (emit-label label)
+    (note-this-location vop :non-local-entry)
+    ;; Get the block saved in UNWIND
+    (inst ldr block (@ csp-tn (* -4 n-word-bytes)))))
+
+#+unwind-to-frame-and-call-vop
 (define-vop (unwind-to-frame-and-call)
   (:args (ofp :scs (descriptor-reg))
          (uwp :scs (descriptor-reg))
-         (function :scs (descriptor-reg) :to :load :target saved-function))
-  (:arg-types system-area-pointer system-area-pointer t)
+         (function :scs (descriptor-reg) :to :load :target saved-function)
+         (bsp :scs (any-reg descriptor-reg))
+         (nsp :scs (any-reg descriptor-reg))
+         (catch-block :scs (any-reg descriptor-reg)))
+  (:arg-types system-area-pointer system-area-pointer t t t t)
   (:temporary (:sc unsigned-reg) temp)
-  (:temporary (:sc descriptor-reg :offset r8-offset) saved-function)
+  (:temporary (:sc descriptor-reg :offset r9-offset) saved-function)
   (:temporary (:sc unsigned-reg :offset r0-offset) block)
   (:temporary (:sc descriptor-reg :offset lexenv-offset) lexenv)
   (:temporary (:scs (interior-reg)) lip)
@@ -259,7 +263,7 @@
     (let ((entry-label (gen-label)))
       ;; Store the function into a non-stack location, since we'll be
       ;; unwinding the stack and destroying register contents before we
-      ;; use it.  It turns out that R8 is preserved as part of the
+      ;; use it.  It turns out that R9 is preserved as part of the
       ;; normal multiple-value handling of an unwind, so use that.
       (move saved-function function)
 
@@ -273,21 +277,22 @@
       (storew temp block unwind-block-uwp-slot)
       (loadw temp ofp sap-pointer-slot other-pointer-lowtag)
       (storew temp block unwind-block-cfp-slot)
-      ;; Don't need to save code at unwind-block-code-slot since
-      ;; it's not going to be used and will be overwritten after the
-      ;; function call
+      (storew bsp block unwind-block-bsp-slot)
+      (storew nsp block unwind-block-nsp-slot)
+      (storew catch-block block unwind-block-current-catch-slot)
 
-      (inst compute-lra temp lip entry-label)
+      (inst adr temp entry-label)
       (storew temp block catch-block-entry-pc-slot)
 
       ;; Run any required UWPs.
       (load-inline-constant tmp-tn '(:fixup unwind :assembly-routine) lip)
       (inst br tmp-tn)
 
-      (emit-return-pc ENTRY-LABEL)
+      (emit-label ENTRY-LABEL)
       (inst mov nargs 0)
 
       (move lexenv saved-function)
 
       (loadw saved-function lexenv closure-fun-slot fun-pointer-lowtag)
       (lisp-jump saved-function lip))))
+
