@@ -12,7 +12,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!C")
+(in-package "SB-C")
 
 ;;;; cost estimation
 
@@ -47,8 +47,7 @@
         0)
       (when (eq type *empty-type*)
         0)
-      (let ((found (cdr (assoc type *backend-type-predicates*
-                               :test #'type=))))
+      (let ((found (backend-type-predicate type)))
         (if found
             (+ (fun-guessed-cost found) (fun-guessed-cost 'eq))
             nil))
@@ -139,7 +138,7 @@
          ;; weakening, so for integer types we simply collapse all
          ;; ranges into one.
          (weaken-integer-type type))
-        #!+sb-unicode
+        #+sb-unicode
         ((csubtypep type (specifier-type 'base-char))
          ;; Don't want to be putting CHARACTERs into BASE-STRINGs.
          (specifier-type 'base-char))
@@ -194,9 +193,9 @@
 (defun maybe-weaken-check (type policy)
   (declare (type ctype type))
   (typecase type
-    ;; Can't do much funcational type checking at run-time
+    ;; Can't do much functional type checking at run-time
     (fun-designator-type
-     (specifier-type '(or function symbol)))
+     (specifier-type 'function-designator))
     (fun-type
      (specifier-type 'function))
     (t
@@ -205,6 +204,12 @@
        (2 (weaken-values-type type))
        (3 type)))))
 
+(defun type-contains-fun-type-p (type)
+  (sb-kernel::map-type (lambda (type)
+                         (when (fun-type-p type)
+                           (return-from type-contains-fun-type-p t)))
+                       type))
+
 ;;; LVAR is an lvar we are doing a type check on and TYPES is a list
 ;;; of types that we are checking its values against. If we have
 ;;; proven that LVAR generates a fixed number of values, then for each
@@ -222,6 +227,9 @@
                        c)
           for diff = (type-difference p cc)
           collect (if (and diff
+                           ;; FUN-TYPE is weakend to FUNCTION, can't
+                           ;; use that for invereted type testes
+                           (not (type-contains-fun-type-p diff))
                            (< (type-test-cost diff)
                               (type-test-cost cc)))
                       (list t diff a)
@@ -346,7 +354,7 @@
                    (remove-if (lambda (spec)
                                 (or (stringp spec)
                                     (typep spec '(cons (eql or)))))
-                              sb!c:+backend-internal-errors+
+                              sb-c:+backend-internal-errors+
                               :key #'car)))
            ;; This is effectively a compact read-only binned hashtable.
            (hashtable (make-array (logior (length entries) 1)
@@ -354,7 +362,7 @@
         (map nil
              (lambda (entry)
                (let* ((canon-type (car entry))
-                      (bucket (mod (sxhash canon-type) (length hashtable))))
+                      (bucket (mod (cl:sxhash canon-type) (length hashtable))))
                  (push entry (svref hashtable bucket))))
              entries)
       hashtable))
@@ -366,7 +374,7 @@
                    (cadr entry)))
            (remove-if-not (lambda (spec)
                             (typep spec '(cons (eql or))))
-                          sb!c:+backend-internal-errors+
+                          sb-c:+backend-internal-errors+
                           :key #'car)))
 (declaim (simple-vector **type-spec-unions-interr-symbols**))
 
@@ -375,9 +383,9 @@
 (defun %interr-symbol-for-union-type-spec (spec)
   (let* ((spec (cdr spec))
          (length (length spec))
-         (bit-map (if (> length sb!vm:n-fixnum-bits)
+         (bit-map (if (> length sb-vm:n-fixnum-bits)
                       (return-from %interr-symbol-for-union-type-spec)
-                      (1- (ash 1 (truly-the (integer 1 #.sb!vm:n-fixnum-bits)
+                      (1- (ash 1 (truly-the (integer 1 #.sb-vm:n-fixnum-bits)
                                             length))))))
     (declare (list spec))
     (loop for entry across **type-spec-unions-interr-symbols**
@@ -390,7 +398,7 @@
                   for position = (position element spec :test #'equal)
                   always position
                   do
-                  (setf (ldb (byte 1 (truly-the (integer 0 (#.sb!vm:n-fixnum-bits))
+                  (setf (ldb (byte 1 (truly-the (integer 0 (#.sb-vm:n-fixnum-bits))
                                                 position))
                              current-map)
                         0))
@@ -401,7 +409,7 @@
   (let ((table **type-spec-interr-symbols**))
     (if (typep spec '(cons (eql or)))
         (%interr-symbol-for-union-type-spec spec)
-        (cdr (assoc spec (svref table (rem (sxhash spec) (length table)))
+        (cdr (assoc spec (svref table (rem (cl:sxhash spec) (length table)))
                     :test #'equal)))))
 #+nil ; some meta-analysis to decide what types should be in "generic/interr"
 (progn
@@ -444,54 +452,39 @@
 ;;; unsupplied. Such checking is impossible to efficiently do at the
 ;;; source level because our fixed-values conventions are optimized
 ;;; for the common MV-BIND case.
-(defun make-type-check-form (types &optional context)
+(defun make-type-check-form (types cast)
   (let ((temps (make-gensym-list (length types))))
-    `(multiple-value-bind ,temps 'dummy
-       ,@(mapcar (lambda (temp %type)
-                   (destructuring-bind (not type-to-check
-                                        type-to-report) %type
-                    (let* ((spec
-                             (let ((*unparse-fun-type-simplify* t))
-                               (type-specifier type-to-check)))
-                           (test (if not `(not ,spec) spec)))
-                      `(unless (typep ,temp ',test)
-                         ,(internal-type-error-call temp
-                                                    (if (fun-designator-type-p type-to-report)
-                                                        ;; Simplify
-                                                        (specifier-type 'callable)
-                                                        type-to-report)
-                                                    context)))))
-                 temps
-                 types)
-       (values ,@temps))))
+    (lambda (dummy)
+      `(multiple-value-bind ,temps ,dummy
+         ,@(mapcar
+            (lambda (temp %type)
+              (destructuring-bind (not type-to-check
+                                   type-to-report) %type
+                (let* ((spec
+                         (let ((*unparse-fun-type-simplify* t))
+                           (type-specifier type-to-check)))
+                       (test (if not `(not ,spec) spec)))
+                  `(unless
+                       ,(with-ir1-environment-from-node cast ;; it performs its own inlining of SATISFIES
+                          (%source-transform-typep temp test))
+                     ,(internal-type-error-call temp
+                                                (if (fun-designator-type-p type-to-report)
+                                                    ;; Simplify
+                                                    (specifier-type 'function-designator)
+                                                    type-to-report)
+                                                (cast-context cast))))))
+            temps
+            types)
+         (values ,@temps)))))
 
 ;;; Splice in explicit type check code immediately before CAST. This
 ;;; code receives the value(s) that were being passed to CAST-VALUE,
 ;;; checks the type(s) of the value(s), then passes them further.
 (defun convert-type-check (cast types)
   (declare (type cast cast) (type list types))
-  (let ((value (cast-value cast))
-        (length (length types)))
-    (filter-lvar value (make-type-check-form types
-                                             (cast-context cast)))
-    (reoptimize-lvar (cast-value cast))
-    (setf (cast-type-to-check cast) *wild-type*)
-    (setf (cast-%type-check cast) nil)
-    (let* ((atype (cast-asserted-type cast))
-           (atype (cond ((not (values-type-p atype))
-                         atype)
-                        ((= length 1)
-                         (single-value-type atype))
-                        (t
-                         (make-values-type
-                          :required (values-type-in atype length)))))
-           (dtype (node-derived-type cast))
-           (dtype (make-values-type
-                   :required (values-type-in dtype length))))
-      (setf (cast-asserted-type cast) atype)
-      (setf (node-derived-type cast) dtype)))
-
-  (values))
+  (filter-lvar (cast-value cast)
+               (make-type-check-form types cast))
+  (setf (cast-%type-check cast) nil))
 
 ;;; Check all possible arguments of CAST and emit type warnings for
 ;;; those with type errors. If the value of USE is being used for a
@@ -526,14 +519,14 @@
         (cond ((and (ref-p use) (constant-p (ref-leaf use)))
                (warn condition
                      :format-control "~:[This~;~:*~A~] is not a ~
-                       ~<~%~9T~:;~/sb!impl:print-type/:~>~% ~S"
+                       ~<~%~9T~:;~/sb-impl:print-type/:~>~% ~S"
                      :format-arguments
                      (list what atype (constant-value (ref-leaf use)))))
               (t
                (warn condition
                      :format-control
-                      "~:[Result~;~:*~A~] is a ~/sb!impl:print-type/, ~
-                       ~<~%~9T~:;not a ~/sb!impl:print-type/.~>"
+                      "~:[Result~;~:*~A~] is a ~/sb-impl:print-type/, ~
+                       ~<~%~9T~:;not a ~/sb-impl:print-type/.~>"
                      :format-arguments (list what dtype atype)))))))
   (values))
 
@@ -566,36 +559,53 @@
 ;;; which may lead to inappropriate template choices due to the
 ;;; modification of argument types.
 (defun generate-type-checks (component)
-  (collect ((casts))
-    (do-blocks (block component)
-      (when (and (block-type-check block)
-                 (not (block-delete-p block)))
-        ;; CAST-EXTERNALLY-CHECKABLE-P wants the backward pass
-        (do-nodes-backwards (node nil block)
-          (when (and (cast-p node)
-                     (cast-type-check node))
-            (cast-check-uses node)
-            (cond ((cast-externally-checkable-p node)
-                   (setf (cast-%type-check node) :external))
-                  (t
-                   ;; it is possible that NODE was marked :EXTERNAL by
-                   ;; the previous pass
-                   (setf (cast-%type-check node) t)
-                   (casts node)))))
-        (setf (block-type-check block) nil)))
-    (dolist (cast (casts))
-      (unless (bound-cast-p cast)
-        (multiple-value-bind (check types) (cast-check-types cast)
-          (ecase check
-            (:simple
-             (convert-type-check cast types))
-            (:too-hairy
-             (let ((*compiler-error-context* cast))
+  (let (generated)
+    (collect ((casts))
+      (do-blocks (block component)
+        (when (and (block-type-check block)
+                   (not (block-delete-p block)))
+          ;; CAST-EXTERNALLY-CHECKABLE-P wants the backward pass
+          (do-nodes-backwards (node nil block)
+            (when (and (cast-p node)
+                       (cast-type-check node))
+              (cast-check-uses node)
+              (cond ((cast-externally-checkable-p node)
+                     (setf (cast-%type-check node) :external))
+                    (t
+                     ;; it is possible that NODE was marked :EXTERNAL by
+                     ;; the previous pass
+                     (setf (cast-%type-check node) t)
+                     (casts node)))))
+          (setf (block-type-check block) nil)))
+      (dolist (cast (casts))
+        (unless (bound-cast-p cast)
+          (multiple-value-bind (check types) (cast-check-types cast)
+            (ecase check
+              (:simple
+               (convert-type-check cast types)
+               (setf generated t))
+              (:too-hairy
                (when (policy cast (>= safety inhibit-warnings))
-                 (compiler-notify
-                  "type assertion too complex to check:~%~
-                    ~/sb!impl:print-type/."
-                  (coerce-to-values (cast-asserted-type cast)))))
-             (setf (cast-type-to-check cast) *wild-type*)
-             (setf (cast-%type-check cast) nil)))))))
-  (values))
+                 (let* ((*compiler-error-context* cast)
+                       (type (cast-asserted-type cast))
+                       (value-type (coerce-to-values type)))
+                   (compiler-notify
+                    "Type assertion too complex to check:~@
+                    ~/sb-impl:print-type/.~a"
+                    type
+                    (cond ((values-type-rest value-type)
+                           (format nil
+                                   "~%It allows an unknown number of values, consider using~@
+                                    ~/sb-impl:print-type/."
+                                   (make-values-type :required (values-type-required value-type)
+                                                     :optional (values-type-optional value-type))))
+                          ((values-type-optional value-type)
+                           (format nil
+                                   "~%It allows a variable number of values, consider using~@
+                                    ~/sb-impl:print-type/."
+                                   (make-values-type :required (append (values-type-required value-type)
+                                                                       (values-type-optional value-type)))))
+                          ("")))))
+               (setf (cast-type-to-check cast) *wild-type*)
+               (setf (cast-%type-check cast) nil)))))))
+    generated))
