@@ -27,31 +27,25 @@
 (in-package "SB-PCL")
 
 (defmacro wrapper-class (wrapper)
-  `(classoid-pcl-class (layout-classoid ,wrapper)))
-
-(declaim (inline make-wrapper-internal))
-(defun make-wrapper-internal (&key (bitmap -1) length classoid)
-  (make-layout classoid :length length :invalid nil
-               :flags +pcl-object-layout-flag+ :bitmap bitmap))
+  `(classoid-pcl-class (wrapper-classoid ,wrapper)))
 
 ;;; This is called in BRAID when we are making wrappers for classes
 ;;; whose slots are not initialized yet, and which may be built-in
-;;; classes. We pass in the class name in addition to the class.
-(defun !boot-make-wrapper (length name &optional class)
+;;; classes.
+(defun !boot-make-wrapper (length name &optional (bitmap +layout-all-tagged+))
   (let ((found (find-classoid name nil)))
     (cond
      (found
-      (unless (classoid-pcl-class found)
-        (setf (classoid-pcl-class found) class))
-      (aver (eq (classoid-pcl-class found) class))
-      (let ((layout (classoid-layout found)))
+      (let ((layout (classoid-wrapper found)))
         (aver layout)
         layout))
      (t
-      (make-wrapper-internal
-       :length length
-       :classoid (make-standard-classoid
-                  :name name :pcl-class class))))))
+       (make-layout (hash-layout-name name)
+                    (make-standard-classoid :name name)
+                    :length length
+                    :flags +pcl-object-layout-flag+
+                    :bitmap bitmap
+                    :invalid nil)))))
 
 ;;; In SBCL, as in CMU CL, the layouts (a.k.a wrappers) for built-in
 ;;; and structure classes already exist when PCL is initialized, so we
@@ -64,46 +58,31 @@
   (cond
     ((or (typep class 'std-class)
          (typep class 'forward-referenced-class))
-     (make-wrapper-internal
-      :length length
-      :classoid
-      (let ((owrap (class-wrapper class)))
-        (cond (owrap
-               (layout-classoid owrap))
-              ((or (*subtypep (class-of class) *the-class-standard-class*)
-                   (*subtypep (class-of class) *the-class-funcallable-standard-class*)
-                   (typep class 'forward-referenced-class))
-               (let ((name (slot-value class 'name)))
-                 (make-standard-classoid :pcl-class class
-                                         :name (and (symbolp name) name))))
-              (t
-               (bug "Got to T branch in ~S" 'make-wrapper))))))
+     (let* ((name)
+            (classoid
+             (let ((owrap (class-wrapper class)))
+               (cond (owrap
+                      (wrapper-classoid owrap))
+                     ((or (*subtypep (class-of class) *the-class-standard-class*)
+                          (*subtypep (class-of class) *the-class-funcallable-standard-class*)
+                          (typep class 'forward-referenced-class))
+                      (setq name (slot-value class 'name))
+                      (make-standard-classoid :pcl-class class
+                                              :name (and (symbolp name) name)))
+                     (t
+                      (bug "Got to T branch in ~S" 'make-wrapper))))))
+        (make-layout (hash-layout-name name) classoid
+                     :bitmap +layout-all-tagged+
+                     :invalid nil :length length :flags +pcl-object-layout-flag+)))
     (t
      (let* ((found (find-classoid (slot-value class 'name)))
-            (layout (classoid-layout found)))
+            (layout (classoid-wrapper found)))
        (unless (classoid-pcl-class found)
          (setf (classoid-pcl-class found) class))
        (aver (eq (classoid-pcl-class found) class))
        (aver layout)
        layout))))
 
-(declaim (inline wrapper-class*))
-(defun wrapper-class* (wrapper)
-  (or (wrapper-class wrapper)
-      ;; FIXME: this branch seems unreachable.
-      ;; It would be nice to eliminate WRAPPER-CLASS* if we can show that it
-      ;; is only a holdover from an earlier way of bootstrapping that resulted
-      ;; in the temporary absence of a PCL-CLASS for some non-standard-class.
-      ;; Certainly no test gets here [changing it to (BUG "got here") worked].
-      ;; Note however that
-      ;;  (CLASSOID-PCL-CLASS (FIND-CLASSOID 'STANDARD-INSTANCE)) => NIL
-      ;; which can be resolved by just ensuring one time that it has a CLASS.
-      ;; And nothing else seems to be problematic.
-      (let ((classoid (layout-classoid wrapper)))
-        (ensure-non-standard-class
-         (classoid-name classoid)
-         classoid))))
-
 ;;; The wrapper cache machinery provides general mechanism for
 ;;; trapping on the next access to any instance of a given class. This
 ;;; mechanism is used to implement the updating of instances when the
@@ -117,9 +96,18 @@
 ;;; SLOT-VALUE-USING-CLASS check the wrapper validity as well. This is
 ;;; done by calling CHECK-WRAPPER-VALIDITY.
 
+;;; This "simple" function hides a horrible inconsistency: we don't have a single
+;;; canonical atomically checkable validity indicator. We carry around invalid layouts
+;;; with nonzero hashes. Why else would there be so may places that can do:
+;;;   (setf (wrapper-invalid layout) nil
+;;; whilst the layout in question already has a "valid" hash?
+;;; It's hard to know what the right thing is, but since internal code uses
+;;; this test to decide how to handle layouts that have been invalidated,
+;;; I don't think we can just change it to examine the hash value
+;;; even though logically that *should* be the canonical test.
 (declaim (inline invalid-wrapper-p))
 (defun invalid-wrapper-p (wrapper)
-  (not (null (layout-invalid wrapper))))
+  (not (null (wrapper-invalid wrapper))))
 
 ;;; The portable code inherited from the ancient PCL sources used an extremely
 ;;; clever mechanism to store a mapping from new layouts ("wrappers") to old.
@@ -168,12 +156,11 @@
 ;;; but also pretty much not correct.
 (defun %invalidate-wrapper (owrapper state nwrapper)
   (aver (member state '(:flush :obsolete)))
-  #+sb-thread (aver (eq (sb-thread:mutex-owner sb-c::**world-lock**)
-                        sb-thread:*current-thread*))
-  (let ((classoid (layout-classoid nwrapper))
+  #+sb-thread (aver (sb-thread:holding-mutex-p sb-c::**world-lock**))
+  (let ((classoid (wrapper-classoid nwrapper))
         (new-previous ())
         (new-state (cons state nwrapper)))
-    (aver (eq (layout-classoid owrapper) classoid))
+    (aver (eq (wrapper-classoid owrapper) classoid))
     ;; First off, a previous call to INVALIDATE-WRAPPER may have
     ;; recorded OWRAPPER as an NWRAPPER to update to. Since OWRAPPER
     ;; is about to be invalid, it no longer makes sense to update to
@@ -185,8 +172,8 @@
     (dolist (weak-pointer (sb-kernel::standard-classoid-old-layouts classoid))
       (let ((previous (weak-pointer-value weak-pointer)))
         (when previous
-          (setf (layout-invalid previous)
-                (cond ((and (eq (car (layout-invalid previous)) :obsolete)
+          (setf (wrapper-invalid previous)
+                (cond ((and (eq (car (wrapper-invalid previous)) :obsolete)
                             (eq state :flush))
                        ;; :obsolete must stay :obsolete,
                        ;; requiring the protocol for obsolete instances.
@@ -198,10 +185,14 @@
 
     ;; FIXME: We are here inside PCL lock, but might someone be
     ;; accessing the wrapper at the same time from outside the lock?
-    (setf (layout-clos-hash owrapper) 0)
-    ;; And shouldn't these assignments be flipped, so that if an observer
-    ;; sees a 0 hash in owrapper, it can find the new layout?
-    (setf (layout-invalid owrapper) new-state)
+    ;; Inform readers of the reason for wrapper invalidity before marking
+    ;; the wrapper as invalid.
+    (setf (wrapper-invalid owrapper) new-state)
+    ;; Ensure that the INVALID slot conveying ancillary data describing the
+    ;; invalidity reason is published before causing the invalid layout trap.
+    (sb-thread:barrier (:write))
+    #+metaspace (setf (layout-clos-hash (wrapper-friend owrapper)) 0)
+    (setf (wrapper-clos-hash owrapper) 0)
     (push (make-weak-pointer owrapper) new-previous)
     ;; This function is called for effect; return value is arbitrary.
     (setf  (sb-kernel::standard-classoid-old-layouts classoid)
@@ -212,12 +203,12 @@
 ;;; (or the names of our callees.)
 (defun check-wrapper-validity (instance)
   (with-world-lock ()
-    (let* ((owrapper (layout-of instance))
-           (state (layout-invalid owrapper)))
+    (let* ((owrapper (wrapper-of instance))
+           (state (wrapper-invalid owrapper)))
       (aver (not (eq state :uninitialized)))
       (cond ((not state)
              owrapper)
-            ((not (layout-for-std-class-p owrapper))
+            ((not (layout-for-pcl-obj-p owrapper))
              ;; Obsolete structure trap.
              (%obsolete-instance-trap owrapper nil instance))
             ((eq t state)
@@ -231,7 +222,7 @@
              ;;    previous call to REGISTER-LAYOUT for a superclass of
              ;;    INSTANCE's class.  See also the comment above
              ;;    FORCE-CACHE-FLUSHES.  Paul Dietz has test cases for this.
-             (let ((class (wrapper-class* owrapper)))
+             (let ((class (wrapper-class owrapper)))
                (%force-cache-flushes class)
                ;; KLUDGE: avoid an infinite recursion, it's still better to
                ;; bail out with an error for server softwares. see FIXME above.
@@ -240,7 +231,7 @@
                ;; Error message here is trying to figure out a bit more about the
                ;; situation, since we don't have anything approaching a test-case
                ;; for the bug.
-               (let ((new-state (layout-invalid (layout-of instance))))
+               (let ((new-state (wrapper-invalid (wrapper-of instance))))
                  (when (eq new-state t)
                    (cerror "Nevermind and recurse." 'bug
                            :format-control "~@<~4IProblem forcing cache flushes. Please report ~
@@ -249,31 +240,31 @@
                                             ~% Wrapper-of: ~S~
                                             ~% Class-wrapper: ~S~%~:@>"
                            :format-arguments (mapcar (lambda (x)
-                                                       (cons x (layout-invalid x)))
+                                                       (cons x (wrapper-invalid x)))
                                                      (list owrapper
-                                                           (layout-of instance)
+                                                           (wrapper-of instance)
                                                            (class-wrapper class)))))))
              (check-wrapper-validity instance))
             ((consp state)
-             (ecase (car state)
-               (:flush
-                (let ((new (cdr state)))
+             (let ((new (the wrapper (cdr state))))
+               (ecase (car state)
+                 (:flush
                   (cond ((std-instance-p instance)
-                         (setf (%instance-layout instance) new))
+                         (setf (%instance-wrapper instance) new))
                         ((fsc-instance-p instance)
-                         (setf (%funcallable-instance-layout instance) new))
+                         (setf (%fun-wrapper instance) new))
                         (t
-                         (bug "unrecognized instance type")))))
-               (:obsolete
-                (%obsolete-instance-trap owrapper (cdr state) instance))))))))
+                         (bug "unrecognized instance type"))))
+                 (:obsolete
+                  (%obsolete-instance-trap owrapper new instance)))))))))
 
 (declaim (inline check-obsolete-instance))
 (defun check-obsolete-instance (instance)
-  (when (invalid-wrapper-p (layout-of instance))
+  (when (invalid-wrapper-p (wrapper-of instance))
     (check-wrapper-validity instance)))
 
 (defun valid-wrapper-of (instance)
-  (let ((wrapper (layout-of instance)))
+  (let ((wrapper (wrapper-of instance)))
     (if (invalid-wrapper-p wrapper)
         (check-wrapper-validity instance)
         wrapper)))
@@ -363,7 +354,7 @@
                   `((class *the-class-t*)
                     (type t))))
          (unless (eq mt t)
-           (setq wrapper (layout-of arg))
+           (setq wrapper (wrapper-of arg))
            (when (invalid-wrapper-p wrapper)
              (setq ,invalid-wrapper-p t)
              (setq wrapper (check-wrapper-validity arg)))
@@ -377,7 +368,7 @@
                     (setf (cdr dfun-wrappers-tail) new-dfun-wrappers-tail)
                     (setf dfun-wrappers-tail new-dfun-wrappers-tail))))
            ,@(when wrappers
-               `((setq class (wrapper-class* wrapper))
+               `((setq class (wrapper-class wrapper))
                  (setq type `(class-eq ,class)))))
          ,@(when wrappers
              `((push wrapper wrappers-rev)
